from structlog import get_logger


class BaseDns(object):
    """
    """

    def __init__(self):
        self.dns_provider_name = None
        if self.dns_provider_name is None:
            raise ValueError(
                'The class attribute dns_provider_name ought to be defined.')

        self.logger = get_logger(__name__).bind(
            dns_provider_name=self.dns_provider_name)

    def log_response(self, response):
        """
        renders a python-requests response as json or as a string
        """
        try:
            response.content.decode('ascii')  # try and trigger a unicode error.
            log_body = response.json()
        except UnicodeError:
            # unicodeError is a subclass of ValueError so we need to capture it first
            log_body = 'Unicode response.'
        except ValueError:
            log_body = response.content
        return log_body

    def create_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        """
        Method that creates/adds a dns TXT record for a domain/subdomain name on
        a chosen DNS provider.

        :param domain_name: :string: The domain/subdomain name whose dns record ought to be
            created/added on a chosen DNS provider.
        :param base64_of_acme_keyauthorization: :string: The value/content of the TXT record that will be
            created/added for the given domain/subdomain

        This method should return None

        Basic Usage:
            If the value of the `domain_name` variable is example.com and the value of
            `base64_of_acme_keyauthorization` is HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld
            Then, your implementation of this method ought to create a DNS TXT record
            whose name is '_acme-challenge' + '.' + domain_name + '.' (ie: _acme-challenge.example.com. )
            and whose value/content is HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld

            Using a dns client like dig(https://linux.die.net/man/1/dig) to do a dns lookup should result
            in something like:
                dig TXT _acme-challenge.example.com
                ...
                ;; ANSWER SECTION:
                _acme-challenge.example.com. 120 IN TXT "HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld"
                _acme-challenge.singularity.brandur.org. 120 IN TXT "9C0DqKC_4MkowIFByHhFaP8u0Zv4z7Wz2IHM91lTKec"
            Optionally, you may also use an online dns client like: https://toolbox.googleapps.com/apps/dig/#TXT/

            Please consult your dns provider on how/format of their DNS TXT records.
            You may also want to consult the cloudflare DNS implementation that is found in this repository.
        """
        self.logger.info('create_dns_record')
        raise NotImplementedError(
            'create_dns_record method must be implemented.')

    def delete_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        """
        Method that deletes/removes a dns TXT record for a domain/subdomain name on
        a chosen DNS provider.

        :param domain_name: :string: The domain/subdomain name whose dns record ought to be
            deleted/removed on a chosen DNS provider.
        :param base64_of_acme_keyauthorization: :string: The value/content of the TXT record that will be
            deleted/removed for the given domain/subdomain

        This method should return None
        """
        self.logger.info('delete_dns_record')
        raise NotImplementedError(
            'delete_dns_record method must be implemented.')



# DNS Provider for AuroRa DNS from the dutch hosting provider pcextreme
# https://www.pcextreme.nl/aurora/dns
# Aurora uses libcloud from apache
# https://libcloud.apache.org/

from libcloud.dns.providers import get_driver
from libcloud.dns.types import Provider, RecordType
import tldextract
from structlog import get_logger
from . import common


class AuroraDns(common.BaseDns):
    """
    Todo: re-organize this class so that we make it easier to mock things out to
    facilitate better tests.
    """

    def __init__(self, AURORA_API_KEY, AURORA_SECRET_KEY):

        self.AURORA_API_KEY = AURORA_API_KEY
        self.AURORA_SECRET_KEY = AURORA_SECRET_KEY
        self.dns_provider_name = 'aurora'

        self.logger = get_logger(__name__).bind(
            dns_provider_name=self.dns_provider_name)

    def create_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        self.logger.info('create_dns_record')

        # delete any prior existing DNS authorizations that may exist already
        self.delete_dns_record(
            domain_name=domain_name,
            base64_of_acme_keyauthorization=base64_of_acme_keyauthorization)

        extractedDomain = tldextract.extract(domain_name)
        domainSuffix = extractedDomain.domain + '.' + extractedDomain.suffix

        if extractedDomain.subdomain is '':
            subDomain = '_acme-challenge'
        else:
            subDomain = '_acme-challenge.' + extractedDomain.subdomain

        cls = get_driver(Provider.AURORADNS)
        driver = cls(key=self.AURORA_API_KEY, secret=self.AURORA_SECRET_KEY)
        zone = driver.get_zone(domainSuffix)
        zone.create_record(
            name=subDomain,
            type=RecordType.TXT,
            data=base64_of_acme_keyauthorization)
        return

    def delete_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        self.logger.info('delete_dns_record')

        extractedDomain = tldextract.extract(domain_name)
        domainSuffix = extractedDomain.domain + '.' + extractedDomain.suffix

        if extractedDomain.subdomain is '':
            subDomain = '_acme-challenge'
        else:
            subDomain = '_acme-challenge.' + extractedDomain.subdomain

        cls = get_driver(Provider.AURORADNS)
        driver = cls(key=self.AURORA_API_KEY, secret=self.AURORA_SECRET_KEY)
        zone = driver.get_zone(domainSuffix)

        records = driver.list_records(zone)
        for x in records:
            if x.name == subDomain and x.type == 'TXT':
                record_id = x.id
                self.logger.info('Found record ' + subDomain + '.' +
                                 domainSuffix + ' with id : ' + record_id + '.')
                record = driver.get_record(zone_id=zone.id, record_id=record_id)
                driver.delete_record(record)
                self.logger.info('Deleted record ' + subDomain + '.' +
                                 domainSuffix + ' with id : ' + record_id + '.')
            else:
                self.logger.info('Record ' + subDomain + '.' + domainSuffix +
                                 ' not found. No record to delete.')

        return



import json
import urlparse

import requests
from structlog import get_logger

from . import common


class CloudFlareDns(common.BaseDns):
    """
    """

    def __init__(
            self,
            CLOUDFLARE_DNS_ZONE_ID,
            CLOUDFLARE_EMAIL,
            CLOUDFLARE_API_KEY,
            CLOUDFLARE_API_BASE_URL='https://api.cloudflare.com/client/v4/'):

        self.CLOUDFLARE_DNS_ZONE_ID = CLOUDFLARE_DNS_ZONE_ID
        self.CLOUDFLARE_EMAIL = CLOUDFLARE_EMAIL
        self.CLOUDFLARE_API_KEY = CLOUDFLARE_API_KEY
        self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL
        self.dns_provider_name = 'cloudflare'
        self.HTTP_TIMEOUT = 65  # seconds

        if CLOUDFLARE_API_BASE_URL[-1] != '/':
            self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL + '/'
        else:
            self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL

        self.logger = get_logger(__name__).bind(
            dns_provider_name=self.dns_provider_name)

    def create_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        self.logger.info('create_dns_record')

        # delete any prior existing DNS authorizations that may exist already
        self.delete_dns_record(
            domain_name=domain_name,
            base64_of_acme_keyauthorization=base64_of_acme_keyauthorization)
        url = urlparse.urljoin(self.CLOUDFLARE_API_BASE_URL,
                               'zones/{0}/dns_records'.format(
                                   self.CLOUDFLARE_DNS_ZONE_ID))
        headers = {
            'X-Auth-Email': self.CLOUDFLARE_EMAIL,
            'X-Auth-Key': self.CLOUDFLARE_API_KEY,
            'Content-Type': 'application/json'
        }
        body = {
            "type": "TXT",
            "name": '_acme-challenge' + '.' + domain_name + '.',
            "content": "{0}".format(base64_of_acme_keyauthorization)
        }
        create_cloudflare_dns_record_response = requests.post(
            url,
            headers=headers,
            data=json.dumps(body),
            timeout=self.HTTP_TIMEOUT)
        self.logger.info(
            'create_cloudflare_dns_record_response',
            status_code=create_cloudflare_dns_record_response.status_code,
            response=self.log_response(create_cloudflare_dns_record_response))

    def delete_dns_record(self, domain_name, base64_of_acme_keyauthorization):
        self.logger.info('delete_dns_record')

        class MockResponse(object):

            def __init__(self, status_code=200, content='mock-response'):
                self.status_code = status_code
                self.content = content
                super(MockResponse, self).__init__()

            def json(self):
                return {}

        delete_dns_record_response = MockResponse()
        headers = {
            'X-Auth-Email': self.CLOUDFLARE_EMAIL,
            'X-Auth-Key': self.CLOUDFLARE_API_KEY,
            'Content-Type': 'application/json'
        }

        dns_name = '_acme-challenge' + '.' + domain_name
        list_dns_payload = {'type': 'TXT', 'name': dns_name}
        list_dns_url = urlparse.urljoin(self.CLOUDFLARE_API_BASE_URL,
                                        'zones/{0}/dns_records'.format(
                                            self.CLOUDFLARE_DNS_ZONE_ID))

        list_dns_response = requests.get(
            list_dns_url,
            params=list_dns_payload,
            headers=headers,
            timeout=self.HTTP_TIMEOUT)

        for i in range(0, len(list_dns_response.json()['result'])):
            dns_record_id = list_dns_response.json()['result'][i]['id']
            url = urlparse.urljoin(self.CLOUDFLARE_API_BASE_URL,
                                   'zones/{0}/dns_records/{1}'.format(
                                       self.CLOUDFLARE_DNS_ZONE_ID,
                                       dns_record_id))
            headers = {
                'X-Auth-Email': self.CLOUDFLARE_EMAIL,
                'X-Auth-Key': self.CLOUDFLARE_API_KEY,
                'Content-Type': 'application/json'
            }
            delete_dns_record_response = requests.delete(
                url, headers=headers, timeout=self.HTTP_TIMEOUT)
            self.logger.info(
                'delete_dns_record_response',
                status_code=delete_dns_record_response.status_code,
                response=self.log_response(delete_dns_record_response))



import os

from setuptools import setup, find_packages
# To use a consistent encoding
import codecs

here = os.path.abspath(os.path.dirname(__file__))
about = {}

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = codecs.open('README.md').read()

with open(os.path.join(here, 'sewer', '__version__.py'), 'r') as f:
    exec (f.read(), about)

setup(
    name=about['__title__'],
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    # The project's main homepage.
    url=about['__url__'],
    # Author details
    author=about['__author__'],
    author_email=about['__author_email__'],
    # Choose your license
    license=about['__license__'],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    # What does your project relate to?
    keywords='letsencrypt',
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=['sewer'],
    packages=find_packages(exclude=['docs', '*tests*']),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'requests', 'structlog', 'pyopenssl', 'cryptography', 'tldextract',
        'apache-libcloud'
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev':
        ['coverage', 'pypandoc', 'twine', 'wheel', 'yapf', 'flake8', 'mock', 'pylint', 'autopep8'],
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
    entry_points={
        'console_scripts': [
            'sewer=sewer.cli:main',
            'sewer-cli=sewer.cli:main',
        ],
    },
)

# python packaging documentation:
# 1. https://python-packaging.readthedocs.io/en/latest/index.html
# 2. https://python-packaging-user-guide.readthedocs.io/tutorials/distributing-packages
# a) pip install wheel twine
# b) pip install -e .
# c) python setup.py sdist
# d) python setup.py bdist_wheel
# e) DONT use python setup.py register and python setup.py upload. They use http
# f) twine upload dist/* -r testpypi
# g) pip install -i https://testpypi.python.org/pypi <package name>
# h) twine upload dist/*   # prod pypi
# i) pip install <package name>



import time
import copy
import json
import base64
import hashlib
import binascii
import urlparse
import textwrap
import platform

import requests
import OpenSSL
import cryptography
from structlog import get_logger

import __version__ as sewer_version


class ACMEclient(object):
    """
    todo: improve documentation.

    usage:
        import sewer
        dns_class = sewer.CloudFlareDns(CLOUDFLARE_DNS_ZONE_ID='random',
                                        CLOUDFLARE_EMAIL='example@example.com',
                                        CLOUDFLARE_API_KEY='nsa-grade-api-key')

        1. to create a new certificate.
        client = sewer.Client(domain_name='example.com',
                              dns_class=dns_class)
        certificate = client.cert()
        certificate_key = client.certificate_key
        account_key = client.account_key

        with open('certificate.crt', 'w') as certificate_file:
            certificate_file.write(certificate)

        with open('certificate.key', 'w') as certificate_key_file:
            certificate_key_file.write(certificate_key)


        2. to renew a certificate:
        with open('account_key.key', 'r') as account_key_file:
            account_key = account_key_file.read()

        client = sewer.Client(domain_name='example.com',
                              dns_class=dns_class,
                              account_key=account_key)
        certificate = client.renew()
        certificate_key = client.certificate_key

    todo:
        - handle exceptions
    """

    def __init__(
            self,
            domain_name,
            dns_class,
            domain_alt_names=[],
            registration_recovery_email=None,
            account_key=None,
            bits=2048,
            digest='sha256',
            ACME_REQUEST_TIMEOUT=65,
            ACME_CHALLENGE_WAIT_PERIOD=8,
            GET_NONCE_URL="https://acme-v01.api.letsencrypt.org/directory",
            ACME_CERTIFICATE_AUTHORITY_URL="https://acme-v01.api.letsencrypt.org",
            ACME_CERTIFICATE_AUTHORITY_TOS='https://letsencrypt.org/documents/LE-SA-v1.2-November-15-2017.pdf',
            ACME_CERTIFICATE_AUTHORITY_CHAIN='https://letsencrypt.org/certs/lets-encrypt-x3-cross-signed.pem'
    ):

        self.logger = get_logger(__name__).bind(
            client_name=self.__class__.__name__)

        self.domain_name = domain_name
        self.dns_class = dns_class
        self.domain_alt_names = domain_alt_names
        self.all_domain_names = copy.copy(self.domain_alt_names)
        self.all_domain_names.insert(0, self.domain_name)
        self.registration_recovery_email = registration_recovery_email
        self.bits = bits
        self.digest = digest
        self.ACME_REQUEST_TIMEOUT = ACME_REQUEST_TIMEOUT
        self.ACME_CHALLENGE_WAIT_PERIOD = ACME_CHALLENGE_WAIT_PERIOD
        self.GET_NONCE_URL = GET_NONCE_URL
        self.ACME_CERTIFICATE_AUTHORITY_URL = ACME_CERTIFICATE_AUTHORITY_URL
        self.ACME_CERTIFICATE_AUTHORITY_TOS = ACME_CERTIFICATE_AUTHORITY_TOS
        self.ACME_CERTIFICATE_AUTHORITY_CHAIN = ACME_CERTIFICATE_AUTHORITY_CHAIN
        self.User_Agent = self.get_user_agent()
        self.certificate_key = self.create_certificate_key()
        self.csr = self.create_csr()
        self.certificate_chain = self.get_certificate_chain()

        if not account_key:
            self.account_key = self.create_account_key()
            self.PRIOR_REGISTERED = False
        else:
            self.account_key = account_key
            self.PRIOR_REGISTERED = True

        self.logger = self.logger.bind(
            sewer_client_name=self.__class__.__name__,
            sewer_client_version=sewer_version.__version__,
            domain_names=self.all_domain_names,
            ACME_CERTIFICATE_AUTHORITY_URL=self.ACME_CERTIFICATE_AUTHORITY_URL)

        # for staging/test, use:
        # GET_NONCE_URL="https://acme-staging.api.letsencrypt.org/directory",
        # ACME_CERTIFICATE_AUTHORITY_URL="https://acme-staging.api.letsencrypt.org"

    def log_response(self, response):
        """
        renders response as json or as a string
        """
        # TODO: use this to handle all response logs.
        try:
            response.content.decode('ascii')  # try and trigger a unicode error.
            log_body = response.json()
        except UnicodeError:
            # certificate has been issued
            # unicodeError is a subclass of ValueError so we need to capture it first
            log_body = 'Response probably contains a certificate.'
        except ValueError:
            log_body = response.content
        return log_body

    def get_user_agent(self):
        # TODO: add the sewer-acme versionto the User-Agent
        return "python-requests/{requests_version} ({system}: {machine}) sewer {sewer_version} ({sewer_url})".format(
            requests_version=requests.__version__,
            system=platform.system(),
            machine=platform.machine(),
            sewer_version=sewer_version.__version__,
            sewer_url=sewer_version.__url__)

    def create_account_key(self):
        self.logger.info('create_account_key')
        return self.create_key()

    def create_certificate_key(self):
        self.logger.info('create_certificate_key')
        return self.create_key()

    def create_key(self, key_type=OpenSSL.crypto.TYPE_RSA):
        key = OpenSSL.crypto.PKey()
        key.generate_key(key_type, self.bits)
        private_key = OpenSSL.crypto.dump_privatekey(
            OpenSSL.crypto.FILETYPE_PEM, key)
        return private_key

    def create_csr(self):
        self.logger.info('create_csr')
        X509Req = OpenSSL.crypto.X509Req()
        X509Req.get_subject().CN = self.domain_name

        if self.domain_alt_names:
            SAN = 'DNS:{0}, '.format(self.domain_name).encode('utf8') + \
                  ', '.join('DNS:' + i for i in self.domain_alt_names).encode('utf8')
        else:
            SAN = 'DNS:{0}'.format(self.domain_name).encode('utf8')

        X509Req.add_extensions([
            OpenSSL.crypto.X509Extension(
                'subjectAltName'.encode('utf8'), critical=False, value=SAN)
        ])
        pk = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM,
                                            self.certificate_key)
        X509Req.set_pubkey(pk)
        X509Req.set_version(2)
        X509Req.sign(pk, self.digest)
        return OpenSSL.crypto.dump_certificate_request(
            OpenSSL.crypto.FILETYPE_ASN1, X509Req)

    def get_certificate_chain(self):
        self.logger.info('get_certificate_chain')
        url = self.ACME_CERTIFICATE_AUTHORITY_CHAIN
        headers = {'User-Agent': self.User_Agent}
        get_certificate_chain_response = requests.get(
            url, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers)
        certificate_chain = get_certificate_chain_response.content.decode(
            'utf8')
        self.logger.info(
            'get_certificate_chain_response',
            status_code=get_certificate_chain_response.status_code)

        if get_certificate_chain_response.status_code not in [200, 201]:
            raise ValueError(
                "Error while getting Acme certificate chain: status_code={status_code} response={response}".
                format(
                    status_code=get_certificate_chain_response.status_code,
                    response=self.log_response(get_certificate_chain_response)))
        elif '-----BEGIN CERTIFICATE-----' and '-----END CERTIFICATE-----' not in get_certificate_chain_response.content:
            raise ValueError(
                "Error while getting Acme certificate chain: status_code={status_code} response={response}".
                format(
                    status_code=get_certificate_chain_response.status_code,
                    response=self.log_response(get_certificate_chain_response)))

        return certificate_chain

    def calculate_safe_base64(self, un_encoded_data):
        return base64.urlsafe_b64encode(un_encoded_data).decode('utf8').rstrip(
            '=')

    def sign_message(self, message):
        self.logger.info('sign_message')
        pk = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM,
                                            self.account_key)
        return OpenSSL.crypto.sign(pk, message.encode('utf8'), self.digest)

    def get_acme_header(self):
        self.logger.info('get_acme_header')
        private_key = cryptography.hazmat.primitives.serialization.load_pem_private_key(
            self.account_key,
            password=None,
            backend=cryptography.hazmat.backends.default_backend())
        public_key_public_numbers = private_key.public_key().public_numbers()

        # private key public exponent in hex format
        exponent = "{0:x}".format(public_key_public_numbers.e)
        exponent = "0{0}".format(exponent) if len(exponent) % 2 else exponent
        # private key modulus in hex format
        modulus = "{0:x}".format(public_key_public_numbers.n)
        header = {
            "alg": "RS256",
            "jwk": {
                "e":
                self.calculate_safe_base64(
                    binascii.unhexlify(exponent.encode('utf8'))),
                "kty":
                "RSA",
                "n":
                self.calculate_safe_base64(
                    binascii.unhexlify(modulus.encode('utf8')))
            }
        }
        return header

    def make_signed_acme_request(self, url, payload):
        self.logger.info('make_signed_acme_request')
        payload64 = self.calculate_safe_base64(
            json.dumps(payload).encode('utf8'))
        protected = self.get_acme_header()

        headers = {'User-Agent': self.User_Agent}
        response = requests.get(
            self.GET_NONCE_URL,
            timeout=self.ACME_REQUEST_TIMEOUT,
            headers=headers)
        nonce = response.headers['Replay-Nonce']
        protected["nonce"] = nonce

        protected64 = self.calculate_safe_base64(
            json.dumps(protected).encode('utf8'))
        signature = self.sign_message(
            message="{0}.{1}".format(protected64, payload64))
        data = json.dumps({
            "protected": protected64,
            "payload": payload64,
            "signature": self.calculate_safe_base64(signature)
        })
        headers = {'User-Agent': self.User_Agent}
        response = requests.post(
            url,
            data=data.encode('utf8'),
            timeout=self.ACME_REQUEST_TIMEOUT,
            headers=headers)
        return response

    def acme_register(self):
        """
        Register using a customers account_key.
        This method should only be called if self.PRIOR_REGISTERED == False
        """
        self.logger.info('acme_register')
        if self.registration_recovery_email:
            payload = {
                "resource": "new-reg",
                "agreement": self.ACME_CERTIFICATE_AUTHORITY_TOS,
                "contact":
                ["mailto:{0}".format(self.registration_recovery_email)]
            }
        else:
            payload = {
                "resource": "new-reg",
                "agreement": self.ACME_CERTIFICATE_AUTHORITY_TOS
            }
        url = urlparse.urljoin(self.ACME_CERTIFICATE_AUTHORITY_URL,
                               '/acme/new-reg')
        acme_register_response = self.make_signed_acme_request(
            url=url, payload=payload)
        self.logger.info(
            'acme_register_response',
            status_code=acme_register_response.status_code,
            response=self.log_response(acme_register_response))

        if acme_register_response.status_code not in [201, 409]:
            raise ValueError(
                "Error while registering: status_code={status_code} response={response}".
                format(
                    status_code=acme_register_response.status_code,
                    response=self.log_response(acme_register_response)))

        return acme_register_response

    def get_challenge(self, domain_name):
        self.logger.info('get_challenge')
        payload = {
            "resource": "new-authz",
            "identifier": {
                "type": "dns",
                "value": domain_name
            }
        }
        url = urlparse.urljoin(self.ACME_CERTIFICATE_AUTHORITY_URL,
                               '/acme/new-authz')
        challenge_response = self.make_signed_acme_request(
            url=url, payload=payload)
        self.logger.info(
            'get_challenge_response',
            status_code=challenge_response.status_code,
            response=self.log_response(challenge_response))

        if challenge_response.status_code != 201:
            raise ValueError(
                "Error requesting for challenges: status_code={status_code} response={response}".
                format(
                    status_code=challenge_response.status_code,
                    response=self.log_response(challenge_response)))

        for i in challenge_response.json()['challenges']:
            if i['type'] == 'dns-01':
                dns_challenge = i
        dns_token = dns_challenge['token']
        dns_challenge_url = dns_challenge['uri']

        return dns_token, dns_challenge_url

    def get_keyauthorization(self, dns_token):
        self.logger.info('get_keyauthorization')
        acme_header_jwk_json = json.dumps(
            self.get_acme_header()['jwk'],
            sort_keys=True,
            separators=(',', ':'))
        acme_thumbprint = self.calculate_safe_base64(
            hashlib.sha256(acme_header_jwk_json.encode('utf8')).digest())
        acme_keyauthorization = "{0}.{1}".format(dns_token, acme_thumbprint)
        base64_of_acme_keyauthorization = self.calculate_safe_base64(
            hashlib.sha256(acme_keyauthorization.encode("utf8")).digest())

        return acme_keyauthorization, base64_of_acme_keyauthorization

    def notify_acme_challenge_set(self, acme_keyauthorization,
                                  dns_challenge_url):
        self.logger.info('notify_acme_challenge_set')
        payload = {
            "resource": "challenge",
            "keyAuthorization": "{0}".format(acme_keyauthorization)
        }
        notify_acme_challenge_set_response = self.make_signed_acme_request(
            dns_challenge_url, payload)
        self.logger.info(
            'notify_acme_challenge_set_response',
            status_code=notify_acme_challenge_set_response.status_code,
            response=self.log_response(notify_acme_challenge_set_response))
        return notify_acme_challenge_set_response

    def check_challenge_status(self, dns_challenge_url,
                               base64_of_acme_keyauthorization, domain_name):
        self.logger.info('check_challenge')
        time.sleep(self.ACME_CHALLENGE_WAIT_PERIOD)
        number_of_checks = 0
        maximum_number_of_checks_allowed = 15
        while True:
            try:
                headers = {'User-Agent': self.User_Agent}
                check_challenge_status_response = requests.get(
                    dns_challenge_url,
                    timeout=self.ACME_REQUEST_TIMEOUT,
                    headers=headers)
                challenge_status = check_challenge_status_response.json()[
                    'status']
                number_of_checks = number_of_checks + 1
                self.logger.info(
                    'check_challenge_status_response',
                    status_code=check_challenge_status_response.status_code,
                    response=self.log_response(check_challenge_status_response),
                    number_of_checks=number_of_checks)
                if number_of_checks > maximum_number_of_checks_allowed:
                    raise StopIteration(
                        "Number of checks done is {0} which is greater than the maximum allowed of {1}.".
                        format(number_of_checks,
                               maximum_number_of_checks_allowed))
            except Exception as e:
                self.logger.info('check_challenge', error=str(e))
                self.dns_class.delete_dns_record(
                    domain_name, base64_of_acme_keyauthorization)
                break
            if challenge_status == "pending":
                time.sleep(self.ACME_CHALLENGE_WAIT_PERIOD)
            elif challenge_status == "valid":
                self.dns_class.delete_dns_record(
                    domain_name, base64_of_acme_keyauthorization)
                break
            else:
                # for any other status, sleep
                time.sleep(self.ACME_CHALLENGE_WAIT_PERIOD)
        return check_challenge_status_response

    def get_certificate(self):
        self.logger.info('get_certificate')
        payload = {
            "resource": "new-cert",
            "csr": self.calculate_safe_base64(self.csr)
        }
        url = urlparse.urljoin(self.ACME_CERTIFICATE_AUTHORITY_URL,
                               '/acme/new-cert')
        get_certificate_response = self.make_signed_acme_request(url, payload)
        self.logger.info(
            'get_certificate_response',
            status_code=get_certificate_response.status_code,
            response=self.log_response(get_certificate_response))

        if get_certificate_response.status_code != 201:
            raise ValueError(
                "Error fetching signed certificate: status_code={status_code} response={response}".
                format(
                    status_code=get_certificate_response.status_code,
                    response=self.log_response(get_certificate_response)))

        base64encoded_cert = base64.b64encode(
            get_certificate_response.content).decode('utf8')
        sixty_four_width_cert = textwrap.wrap(base64encoded_cert, 64)
        certificate = '\n'.join(sixty_four_width_cert)

        pem_certificate = """-----BEGIN CERTIFICATE-----\n{0}\n-----END CERTIFICATE-----\n""".format(
            certificate)
        pem_certificate_and_chain = pem_certificate + self.certificate_chain
        return pem_certificate_and_chain

    def just_get_me_a_certificate(self):
        self.logger.info('just_get_me_a_certificate')
        if not self.PRIOR_REGISTERED:
            self.acme_register()
        for domain_name in self.all_domain_names:
            # NB: this means we will only get a certificate; self.get_certificate()
            # if all the SAN succed the following steps
            dns_token, dns_challenge_url = self.get_challenge(domain_name)
            acme_keyauthorization, base64_of_acme_keyauthorization = self.get_keyauthorization(
                dns_token)
            self.dns_class.create_dns_record(domain_name,
                                             base64_of_acme_keyauthorization)
            self.notify_acme_challenge_set(acme_keyauthorization,
                                           dns_challenge_url)
            self.check_challenge_status(
                dns_challenge_url, base64_of_acme_keyauthorization, domain_name)
        certificate = self.get_certificate()

        return certificate

    def cert(self):
        """
        convenience method to get a certificate without much hassle
        """
        return self.just_get_me_a_certificate()

    def certificate(self):
        """
        convenience method to get a certificate without much hassle
        """
        return self.just_get_me_a_certificate()

    def renew(self):
        """
        renews a certificate.
        A renewal is actually just getting a new certificate.
        An issuance request counts as a renewal if it contains the exact same set of hostnames as a previously issued certificate.
            https://letsencrypt.org/docs/rate-limits/
        """
        return self.just_get_me_a_certificate()
